# from cc_faice.agent.main import main


if __name__ == '__main__':
    # exit(main())
    raise DeprecationWarning('faice agent red is deprecated. Use faice exec with ccfaice execution engine instead.')
